import numpy as np
from typing import Union
from scipy import integrate
import matplotlib.pyplot as plt

# use alpha, b, c
def prob(ability: Union[list, np.ndarray, float], 
         params: Union[list, np.ndarray])  -> np.ndarray:
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).T, (1, items))

        params = np.tile(params, num_abilities).reshape(
            (3, num_abilities, items))
        return params[2] + (1 - params[2]) *  (np.exp(np.exp(params[0]) * (ability - params[1])) / (1 + np.exp(np.exp(params[0]) * (ability - params[1]))))

def prob2PM(ability: Union[list, np.ndarray, float],
            params: Union[list, np.ndarray]) -> np.ndarray:
    items = np.array(params).shape[1]
    num_abilities = 1
    if np.array(ability).shape != ():
        num_abilities = np.array(ability).shape[0]
    ability = np.tile(np.array([ability]).T, (1, items))

    params = np.tile(params, num_abilities).reshape(
        (3, num_abilities, items))
    return np.exp(np.exp(params[0]) * (ability - params[1])) / (1 + np.exp(np.exp(params[0]) * (ability - params[1])))

def plot_item_curve(p):
    ability = np.linspace(-10, 10, 100)
    for i in range(p.shape[1]):
        plt.plot(ability, prob(ability, p[:,i].reshape(p.shape[0], 1)), label=f'Question{i+1}')        
    plt.legend()
    plt.show()





# function params
num_nodes = 4
quads_range = 8
data = [[1, 1, 1, 1, 1, 0, 1, 1, 1, 1],
        [0, 0, 1, 0, 0, 0, 0, 1, 0, 1],
        [1, 1, 0, 1, 0, 1, 0, 0, 1, 1],
        [1, 0, 0, 0, 1, 0, 0, 0, 1, 1],
        [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],] # rows are subjects, columns are items

NALT = 5 # number of alternatives, pg 188. or use p instead of 1/nalt
priori_weight_c = 20

integr_range = 10 # use

# priors
alpha_mean = 0
alpha_var = np.sqrt(0.5) # page 187
b_mean = 0
b_var = 1
ALP = priori_weight_c * (1 / NALT) + 1
BET =  priori_weight_c * (1 - 1 / NALT) + 1
orig_params = np.array([0, 0, 1/NALT])

data = np.array(data)
subjects, items = data.shape
params = np.tile(orig_params, items).reshape((items, 3)).T

i = 0
# # for iteration in range(1):
# #     for i in range(items):


# intergal over likelihood of ability for each subject
# denominator in eq 7.14
integ_likelihood = []
for j in range(subjects):
    likelihood = lambda ability: np.prod(prob(ability, params)[0] ** data[j]) * np.prod((1 - prob(ability, params)[0]) ** (1 - data[j]))
    ability = -10
    print(prob(ability, params)[0] ** data[j])
    print((1 - prob(ability, params)[0]) ** (1 - data[j]))
    print(prob(ability, params)[0] ** data[j] * (1 - prob(ability, params)[0]) ** (1 - data[j]))
    print()
    x = np.linspace(-100, 100, 100)
    y = [likelihood(x_i) for x_i in x]
    plt.plot(x, y)
    plt.show()
    # integ_likelihood.append(integrate.quad(likelihood, -10, 10)[0])
integ_likelihood = np.array(integ_likelihood)
quit()

# def likelihood(j, ability):
#     P = prob(ability, params)[0]
#     return np.prod(P ** data[j] * (1 - P) ** (1 - data[j]))

# px_given_u = lambda ability: np.array([likelihood(j, ability) for j in range(subjects)]) / integ_likelihood
# f = lambda ability: np.sum(px_given_u(ability))
# r = lambda i, ability: np.sum(px_given_u(ability) * data[:,i])

# W = lambda i, ability: (prob2PM(ability, params[:,i].reshape(3,1))[0][0] * (1 - prob2PM(ability, params[:,i].reshape(3,1))[0][0])) / \
# (prob(ability, params[:,i].reshape(3,1))[0][0] * (1 - prob(ability, params[:,i].reshape(3,1))[0][0]))


# L1_integrand = lambda ability: (r(i, ability) - f(ability) * prob(ability, params[:,i].reshape(3,1))[0][0]) * W(i, ability) * (ability - params[1][i])
# L1 = np.exp(params[0][i]) * (1 - params[2][i]) * integrate.quad(L1_integrand, -10, 10)[0] - (params[0][i] - alpha_mean) / alpha_var

# L2_integrand = lambda ability: (r(i, ability) - f(ability) * prob(ability, params[:,i].reshape(3,1))[0][0]) * W(i, ability)
# L2 = -np.exp(params[0][i]) * (1 - params[2][i]) * integrate.quad(L2_integrand, -10, 10)[0]  - (params[1][i] - b_mean) / b_var

# L3_integrand = lambda ability: (r(i, ability) - f(ability) * prob(ability, params[:,i].reshape(3,1))[0][0]) / prob(ability, params[:,i].reshape(3,1))[0][0]
# L3 = 1 / (1 - params[2][i]) * integrate.quad(L3_integrand, -10, 10)[0] + (ALP - 2) / params[2][i] - (BET - 2) / (1 - params[2][i])

# # optimize all of this. P is often reused, for example
# # might also be more readible if we add more lambda functions
# lamb11_integrand = lambda ability: f(i, ability) * (ability - params[1][i]) ** 2 * ((prob(ability, params[:,i].reshape(3,1))[0][0] - params[2][i]) / (1 - params[2][i])) ** 2 * ((1 - prob(ability, params[:,i].reshape(3,1))[0][0]) / (prob(ability, params[:,i].reshape(3,1))[0][0]))
# lamb11 = -np.exp(params[0][i]) ** 2 * integrate.quad(lamb11_integrand, -10, 10)[0] - 1 / alpha_var

# lamb22_integrand = lambda ability: f(i, ability) * ((prob(ability, params[:,i].reshape(3,1))[0][0] - params[2][i]) / (1 - params[2][i])) ** 2 * ((1 - prob(ability, params[:,i].reshape(3,1))[0][0]) / (prob(ability, params[:,i].reshape(3,1))[0][0]))
# lamb22 = -np.exp(params[0][i]) ** 2 * integrate.quad(lamb22_integrand, -10, 10)[0] - 1 / b_var

# lamb33_integrand = lambda ability: f(i, ability) * (1 / prob(ability, params[:,i].reshape(3,1))[0][0] - 1) ** 2 * ((1 - prob(ability, params[:,i].reshape(3,1))[0][0]) / (prob(ability, params[:,i].reshape(3,1))[0][0])) 
# lamb33 = -1 * integrate.quad(lamb33_integrand, -10, 10)[0] - (ALP - 2) / params[2][i] ** 2 - (BET - 2) / (1 - params[2][i]) ** 2

# lamb12_integrand = lambda ability: f(i, ability) * (ability - params[1][i]) * ((prob(ability, params[:,i].reshape(3,1))[0][0] - params[2][i]) / (1 - params[2][i])) ** 2 * ((1 - prob(ability, params[:,i].reshape(3,1))[0][0]) / (prob(ability, params[:,i].reshape(3,1))[0][0]))
# lamb12 = np.exp(params[0][i]) ** 2 * integrate.quad(lamb12_integrand, -10, 10)[0]

# lamb13_integrand = lambda ability: f(i, ability) * (ability - params[1][i]) * ((prob(ability, params[:,i].reshape(3,1))[0][0] - params[2][i]) / (1 - params[2][i])**2) * ((1 - prob(ability, params[:,i].reshape(3,1))[0][0]) / (prob(ability, params[:,i].reshape(3,1))[0][0]))
# lamb13 = -np.exp(params[0][i]) ** 2 * integrate.quad(lamb13_integrand, -10, 10)[0]

# lamb23_integrand = lambda ability: f(i, ability) * ((prob(ability, params[:,i].reshape(3,1))[0][0] - params[2][i]) / (1 - params[2][i])**2) * ((1 - prob(ability, params[:,i].reshape(3,1))[0][0]) / (prob(ability, params[:,i].reshape(3,1))[0][0]))
# lamb23 = np.exp(params[0][i]) * integrate.quad(lamb13_integrand, -10, 10)[0]

# cross_deriv_mat = np.array([[lamb11, lamb12, lamb13], [lamb12, lamb22, lamb23], [lamb13, lamb23, lamb33]])
# deriv_mat = np.array([[L1], [L2], [L3]])

# mat_inv = np.linalg.inv(cross_deriv_mat)
# # est = est - np.matmul(mat_inv, deriv_mat)
        


from pathlib import Path
import sys
if str(Path(__file__).parent) not in sys.path:
    sys.path.append(str(Path(__file__).parent))
from openirt.item_model import ItemModel
from typing import Union
import numpy as np
from pl2 import PL2

class PL1(ItemModel):
    def prob(self, ability: Union[list, np.ndarray, float], params: Union[list, np.ndarray]) -> np.ndarray:
        """
        Calculate the probability of a correct response given the ability and item parameters.

        Args:
            ability: A list, numpy array, or float representing the ability of the individual.
            params: A list or numpy array containing the item parameters.

        Returns:
            A numpy array representing the probabilities of correct responses.
        """
        return PL2().prob(ability, [params[0], np.ones(len(params[0]))])

    def estimate_parameters(self, ability: Union[list, np.ndarray], result: Union[list, np.ndarray],
                            sigm_orig=0, end=0.0000001, eps=0.0001) -> np.ndarray:
        """
        Estimate the parameters of the model given the abilities and item responses.

        Args:
            ability: A list or numpy array containing the abilities of individuals.
            result: A list or numpy array representing the item responses.
            sigm_orig: The initial value for the parameter sigm.
            end: The convergence threshold.
            eps: The threshold for detecting a near-singular matrix.

        Returns:
            A numpy array representing the estimated parameters.
        """
        ability = np.array(ability)
        result = np.array(result)
        est = sigm_orig
        prev_est = est + 2 * end
        while abs(est - prev_est) > end:
            P = self.prob(ability, [[est]]).transpose()[0]
            L1 = np.sum(result - P)
            L2 = -np.sum(P * (1 - P))

            if abs(L2) < eps:
                break

            prev_est = est
            est = est - L1 / L2
        return np.array([est])

    def estimate_ability(self, params: Union[list, np.ndarray], results: Union[list, np.ndarray],
                         end=0.00000001, eps=0.01) -> np.ndarray:
        """
        Estimate the ability of an individual given the item parameters and responses.

        Args:
            params: A list or numpy array representing the item parameters.
            results: A list or numpy array representing the item responses.
            end: The convergence threshold.
            eps: The threshold for detecting a near-singular matrix.

        Returns:
            A numpy array representing the estimated ability.
        """
        return PL2().estimate_ability([params[0], np.ones(len(params[0]))], results, end, eps)



from item_model import ItemModel
import numpy as np
from scipy.stats import norm
from typing import Union

class NormModel():
    
    def prob(self, ability: Union[list, np.ndarray, float], params: Union[list, np.ndarray]) -> np.ndarray:
        """_summary_
        takes array of a abilities (n*1) and array of params (2*m) and returns prob (n*m)

        Args:
            ability (Union[list, np.ndarray, float]): _description_
            params (Union[list, np.ndarray]): _description_

        Returns:
            np.ndarray: _description_
        """
        ability = np.array(ability)
        params = np.array(params)
        z = np.matmul(ability[:, np.newaxis], [params[1]])
        z += np.tile(params[0], (len(ability), 1))
        return norm.cdf(z)

    
    def norm_density(self, ability: Union[list, np.ndarray, float], params: Union[list, np.ndarray]) -> np.ndarray:
        ability = np.array(ability)
        params = np.array(params)
        z = np.matmul(ability[:, np.newaxis], [params[1]])
        z += np.tile(params[0], (len(ability), 1))
        return norm.pdf(z)
        
    def estimate_parameters(self, 
                            ability: Union[list, np.ndarray], 
                            result: Union[list, np.ndarray], 
                            sigm_orig=0,
                            lamb_orig=1, 
                            end=0.0000001,  
                            eps=0.1) -> np.ndarray:
        ability = np.array(ability)
        result = np.array(result)
        est = [sigm_orig, lamb_orig]
        prev_est = [0, 0]
        while abs(est[0] - prev_est[0]) > end or abs(est[1] - prev_est[1]) > end:
            P = self.prob(
                ability, [[est[0]], [est[1]]]).transpose()[0]
            h = self.norm_density(ability, [[est[0]], [est[1]]]).transpose()[0]
            # break
            W = h**2 / (P * (1 - P))
            L11 = -np.sum(W)
            L12 = -np.sum(ability * W)
            L22 = -np.sum(ability**2 * W)
            L = np.array([[L11, L12], [L12, L22]])

            if abs(np.linalg.det(L)) < eps:
                break
            L_inv = np.linalg.inv(L)

            v = (result - P) / h
            obs_mat = np.array([np.sum(W * v), np.sum(W * v * ability)])
            prev_est = est
            est = est - np.matmul(L_inv, obs_mat)
        return est
    
    def estimate_ability(self, params: Union[list, np.ndarray], 
                         results: Union[list, np.ndarray],
                         end=0.00000001,
                         eps=0.01) -> np.ndarray:
        est = 0.5
        prev_est = 0
        while abs(est - prev_est) > end:
            P = self.prob(est, params)[0]
            W = (1 - P) * P
            h = self.norm_density(est, params)[0]
            prev_est = est
            denom = np.sum(params[1]**2 * W)
            if denom < eps or np.any(W < eps):
                break
            est = est + (np.sum(params[1] * W  * ((results - P)/ h)) / denom)
        return est
    
    def convert_param_form(self, params):
        return np.array([-params[0] / params[1], 1 / params[1]])

if __name__ == "__main__":
    nm = NormModel()
    results = np.array([0, 1, 1, 0, 1, 1, 1, 1, 1])
    abilities = [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]
    
    params = nm.estimate_parameters(abilities, results)
    nm.plot_item_curve(np.array([params]).T)


import argparse
from irt_instance import IRTInstance

parser = argparse.ArgumentParser(prog='openirt',
                                 description='Item Response Theory package CLI')

parser.add_argument('action', type=str, choices=['train', 'est-item', 'est-ability'], help='The action to perform')
parser.add_argument('results', type=str, help='The path to the results file')
parser.add_argument('--item_params', type=str, help='The path to the item parameters file')
parser.add_argument('--ability', type=str, help='The path to the ability parameters file')
parser.add_argument('--model', type=str, choices=['1pl', '2pl', '3pl', 'norm'],
                    help='The IRT model to use', default='2pl')
parser.add_argument('--row', type=str, default='None', choices=['item', 'subject'], help='The attribute associated with the rows')
parser.add_argument('--col', type=str, default='None', choices=['item', 'subject'] , help='The attribute associated with the columns')
parser.add_argument('--labeled', default=False, action='store_true', help='Specify if the items or subjects are labeled')
parser.add_argument('--saveICC', default=False, action='store_true', help='Save the ICC values to a file')

# Parse the command line arguments
args = parser.parse_args()
if args.row == args.col and args.row != 'None':
    parser.error('Rows and columns must be different dimensions.')

irt_model = IRTInstance(results_file = args.results, 
                        model=args.model, 
                        labeled=args.labeled, 
                        item_params_file=args.item_params, 
                        ability_file=args.ability,
                        row=args.row,
                        col=args.col)

if args.action == 'train':
    irt_model.train_model()
    irt_model.save_item_params()
    irt_model.save_abilities()

if args.action == 'est-item':
    irt_model.estimate_item_params()
    irt_model.save_item_params()
    
if args.action == 'est-ability':
    irt_model.estimate_abilities()
    irt_model.save_abilities()
    
if args.saveICC:
    irt_model.saveICC()


from abc import ABC, abstractmethod
from typing import Union
import numpy as np
import matplotlib.pyplot as plt


class ItemModel():
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def prob(self, ability: Union[list, np.ndarray, float], params: Union[list, np.ndarray]) -> np.ndarray:
        pass
    
    @abstractmethod
    def estimate_parameters(self, 
                            ability: Union[list, np.ndarray], 
                            result: Union[list, np.ndarray]) -> np.ndarray:
        pass
    
    @abstractmethod
    def estimate_ability(self, params: Union[list, np.ndarray], results: Union[list, np.ndarray]) -> np.ndarray:
        pass
    
    def train_em_mle(self, data) -> tuple:
        data = np.array(data)
        subjects, items = data.shape
        
        params = np.array([np.full(items, -1), np.full(items, 1)])
        for j in range(20):
            abilities = np.array([self.estimate_ability(params, data[i]) for i in range(subjects)])
            params = np.array([self.estimate_parameters(abilities, data[:,i]) for i in range(items)]).T
        return abilities, params
    
    def plot_item_curve(self, p):
        ability = np.linspace(-10, 10, 100)
        for i in range(p.shape[1]):
            prob = self.prob(ability, p[:,i].reshape(p.shape[0], 1))
            plt.plot(ability, self.prob(ability, p[:,i].reshape(p.shape[0], 1)), label=f'Question{i+1}')        
        plt.legend()
        plt.show()
    


from pathlib import Path
import sys
if str(Path(__file__).parent) not in sys.path:
    sys.path.append(str(Path(__file__).parent))
from item_model import ItemModel
from typing import Union
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize


class PL3(ItemModel):
    def prob(self, ability: Union[list, np.ndarray, float], params: Union[list, np.ndarray]) -> np.ndarray:
        """
        Calculate the probability of a correct response given the ability and item parameters.

        Args:
            ability: A list, numpy array, or float representing the ability of the individual.
            params: A list or numpy array containing the item parameters.

        Returns:
            A numpy array representing the probabilities of correct responses.
        """
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape((3, num_abilities, items))
        return params[2] + (1 - params[2]) / (1 + np.exp(-params[0] * (ability - params[1])))

    def prob2PM(self, ability: Union[list, np.ndarray, float], params: Union[list, np.ndarray]) -> np.ndarray:
        """
        Calculate the probability of a correct response using the 2-parameter logistic model.

        Args:
            ability: A list, numpy array, or float representing the ability of the individual.
            params: A list or numpy array containing the item parameters.

        Returns:
            A numpy array representing the probabilities of correct responses using the 2-parameter logistic model.
        """
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape((3, num_abilities, items))
        return 1 / (1 + np.exp(-params[0] * (ability - params[1])))

    def estimate_parameters(self, ability: Union[list, np.ndarray], result: Union[list, np.ndarray],
                            a_orig=1, b_orig=0.1, c_orig=0.1, end=0.0000001, eps=0.1) -> np.ndarray:
        """
        Estimate the parameters of the model given the abilities and item responses.

        Args:
            ability: A list or numpy array containing the abilities of individuals.
            result: A list or numpy array representing the item responses.
            a_orig: The initial value for parameter a.
            b_orig: The initial value for parameter b.
            c_orig: The initial value for parameter c.
            end: The convergence threshold.
            eps: The threshold for detecting a near-singular matrix.

        Returns:
            A numpy array representing the estimated parameters.
        """
        ability = np.array(ability)
        result = np.array(result)
        est = np.array([a_orig, b_orig, c_orig])
        prev_est = est + 2 * end
        while np.all(np.abs(prev_est - est) > end):
            P = self.prob(ability, est.reshape((3, 1))).T[0]
            Q = 1 - P
            P_2pm = self.prob2PM(ability, est.reshape((3, 1))).T[0]

            L11 = -np.sum((ability - est[1]) ** 2 * P * Q * (P_2pm / P) ** 2)
            L12 = np.sum(est[0] * (ability - est[1]) * P * Q * (P_2pm / P))
            L13 = -np.sum((ability - est[1]) * (Q / (1 - est[2])) * (P_2pm / P))

            L22 = -est[0] ** 2 * np.sum(P * Q * (P_2pm / P))
            L23 = np.sum(est[0] * (Q / (1 - est[2])) * (P_2pm / P))

            L33 = -np.sum((Q / (1 - est[2])) / (P - est[2]) * (P_2pm / P))

            L = np.array([[L11, L12, L13], [L12, L22, L23], [L13, L23, L33]])

            if abs(np.linalg.det(L)) < eps:
                break
            L_inv = np.linalg.inv(L)

            L1 = np.sum((result - P) * (ability - est[1]) * (P_2pm / P))
            L2 = -est[0] * np.sum((result - P) * (P_2pm / P))
            L3 = np.sum((result - P) / (P - est[2]) * (P_2pm / P))
            obs_mat = np.array([L1, L2, L3])
            prev_est = est
            est = est - np.matmul(L_inv, obs_mat)
        return est

    def estimate_ability(self, params: Union[list, np.ndarray], results: Union[list, np.ndarray],
                         end=0.00000001, eps=0.01) -> np.ndarray:
        """
        Estimate the ability of an individual given the item parameters and responses.

        Args:
            params: A list or numpy array representing the item parameters.
            results: A list or numpy array representing the item responses.
            end: The convergence threshold.
            eps: The threshold for detecting a near-singular matrix.

        Returns:
            A numpy array representing the estimated ability.
        """
        params = np.array(params)
        results = np.array(results)
        est = 0.5
        prev_est = 0
        while abs(est - prev_est) > end:
            P = self.prob(est, params)[0]
            Q = 1 - P
            P_2pm = self.prob2PM(est, params)[0]
            W = P_2pm * (1 - P_2pm)

            denom = - np.sum(params[0] ** 2 * W * (P_2pm / P) ** 2)

            if abs(denom) < eps:
                break
            num = np.sum(params[0] * W * ((results - P) / (P * Q)) * (P_2pm / P))
            prev_est = est
            est = est + (num / denom)
        return est

    def plot_item_curve(self, p):
        """
        Plot the item response curves for given item parameters.

        Args:
            p: A numpy array representing the item parameters.
        """
        ability = np.linspace(-4, 4, 100)
        for i in range(p.shape[1]):
            plt.plot(ability, self.prob(ability, p[:, i].reshape(p.shape[0], 1)), label=f'Question{i + 1}')
        plt.legend()
        plt.show()




import os
import pandas as pd
# from pl1 import PL1
from pl2 import PL2
# from pl3 import PL3
from norm_model import NormModel

class IRTInstance():
    def __init__(self, 
             results_file: str,
             model: str,
             labeled: bool = True,
             item_params_file: str = None,
             ability_file: str = None,
             row: str = None,
             col: str = None) -> None:
        
        # Read the results file
        self.results = self.read_file(results_file, labeled, labeled)
        if not self.isDichotomous(self.results):
            raise ValueError('Results file must be a binary table.')
        self.setDimensions(row, col)
        
        # Read the item parameters file
        if item_params_file:
            self.item_params = self.read_file(item_params_file, rowLabel=False, colLable=labeled)
        else:
            self.item_params = None
        self.check_item_params()
            
        # Read the ability file
        if ability_file:
            self.abilities = self.read_file(ability_file, rowLabel=False, colLable=labeled)
        else:
            self.abilities = None
        self.check_subjects()
        
        self.set_model(model)
        
    def set_model(self, model: str) -> None:
        if model == '1pl':
            self.model = PL1()
        elif model == '2pl':
            self.model = PL2()
        elif model == '3pl':
            self.model = PL3()
        elif model == 'norm':
            self.model = NormModel()
        else:
            raise ValueError('Model must be one of 1pl, 2pl, 3pl, or norm.')
        
    def read_file(self, file_name: str, rowLabel: bool, colLable: bool):
        if not os.path.isfile(file_name):
            raise FileNotFoundError(f"{file_name} not found")
        
        frmat = file_name.split(".")[-1]
        if frmat == "csv":
            if rowLabel:
                index_col = 0
            else:
                index_col = None
            if colLable:
                header = 0
            else:
                header = None
            return pd.read_csv(file_name, index_col=index_col, header=header)
        elif frmat == "json":
            data = self.read_json(file_name)
        elif frmat == "xlsx":
            data = self.read_xlsx(file_name)
        else:
            raise ValueError('Incorrect file format. Only csv, json, and xlsx files are allowed.')
                
    def isDichotomous(self, data) -> bool:
        # Check if any value in the table is not 0 or 1
        invalid_values = (data != 0) & (data != 1)
        return not invalid_values.any().any()
        
    def setDimensions(self, row: str, col: str) -> None:
        if row == 'item' or col == 'subject':
            self.items = self.results.index
            self.subjects = self.results.columns
        else:
            self.items = self.results.columns
            self.subjects = self.results.index
            
    def check_subjects(self):
        if self.abilities is not None:
            if not self.subjects.equals(self.abilities.columns):
                raise ValueError('Subjects in results file and ability file must be the same.')
            
    def check_item_params(self):
        if self.item_params is not None:
            if not self.items.equals(self.item_params.columns):
                raise ValueError('Items in results file and item parameters file must be the same.')
    
    def train_model(self):
        # print(self.results)
        abilities, params = self.model.train_em_mle(self.results)
        
        # self.model.plot_item_curve(params)
        pass
    
    def estimate_item_params(self):
        pass
    
    def estimate_abilities(self):
        pass
    
    def save_item_params(self, dir: str = './'):
        pass
    
    def save_abilities(self, dir: str = './'):
        pass
        
    def saveICC(self, dir: str):
        pass
    